#include <linux/mm.h>
#include <asm/cachectl.h>
#include <asm/mipsconfig.h>
#include <asm/page.h>
#include <asm/pgtable.h>

/*
 * Initialize new page directory with pointers to invalid ptes
 */
void pgd_init(unsigned long page)
{
	unsigned long dummy1, dummy2;

	/*
	 * We generate dirty lines in the datacache, overwrite these lines
	 * with zeros and then flush the cache.  Sounds horribly complicated
	 * but is just a trick to avoid unnecessary loads of from memory
	 * and uncached stores which are very expensive.
	 * FIXME:  This is the same like the R4000 version.  We could do some
	 * R10000 trickery using caching mode "uncached accelerated".
	 */
	__asm__ __volatile__(
		".set\tnoreorder\n\t"
		".set\tnoat\n\t"
		".set\tmips3\n\t"
		"dsll32\t$1,%2,0\n\t"
		"dsrl32\t%2,$1,0\n\t"
		"or\t%2,$1\n"
		"1:\t"
#ifndef CONF_DISABLE_KSEG0_CACHING
		"cache\t%5,(%0)\n\t"
#endif
		"sd\t%2,(%0)\n\t"
		"sd\t%2,8(%0)\n\t"
#ifndef CONF_DISABLE_KSEG0_CACHING
		"cache\t%5,16(%0)\n\t"
#endif
		"sd\t%2,16(%0)\n\t"
		"sd\t%2,24(%0)\n\t"
#ifndef CONF_DISABLE_KSEG0_CACHING
		"cache\t%5,32(%0)\n\t"
#endif
		"sd\t%2,32(%0)\n\t"
		"sd\t%2,40(%0)\n\t"
#ifndef CONF_DISABLE_KSEG0_CACHING
		"cache\t%5,48(%0)\n\t"
#endif
		"sd\t%2,48(%0)\n\t"
		"sd\t%2,56(%0)\n\t"
		"subu\t%1,1\n\t"
		"bnez\t%1,1b\n\t"
		"addiu\t%0,64\n\t"
		".set\tmips0\n\t"
		".set\tat\n\t"
		".set\treorder"
		:"=&r" (dummy1),
		 "=&r" (dummy2)
		:"r" (((unsigned long) invalid_pte_table - PAGE_OFFSET) |
		       _PAGE_TABLE),
		 "0" (page),
		 "1" (PAGE_SIZE/(sizeof(pmd_t)*16)),
		 "i" (Create_Dirty_Excl_D)
		:"$1");
	/*
	 * Now force writeback to ashure values are in the RAM.
	 */
	sys_cacheflush((void *)page, PAGE_SIZE, DCACHE);
}
